@@ -28,8 +28,6 @@
 /* Initial height of the picture gallery. */
 #define GALLERY_INITIAL_HEIGHT 150
 
-#define UNICODE_ZERO_WIDTH_SPACE "\xe2\x80\x8b"
-
 static void
 composer_setup_charset_menu (EMsgComposer *composer)
 {
@@ -141,6 +139,7 @@ e_composer_private_constructed (EMsgComposer *composer)
 	priv->dnd_is_uri = FALSE;
 	priv->check_if_signature_is_changed = FALSE;
 	priv->ignore_next_signature_change = FALSE;
+	priv->dnd_history_saved = FALSE;
 
 	priv->focused_entry = NULL;
 
@@ -276,7 +275,7 @@ e_composer_private_constructed (EMsgComposer *composer)
 	g_free (gallery_path);
 
 	e_signal_connect_notify_swapped (
-		view, "notify::mode",
+		view, "notify::html-mode",
 		G_CALLBACK (composer_update_gallery_visibility), composer);
 
 	g_signal_connect_swapped (
@@ -483,25 +482,15 @@ e_composer_get_default_charset (void)
 
 	charset = g_settings_get_string (settings, "composer-charset");
 
-	/* See what charset the mailer is using.
-	 * XXX We should not have to know where this lives in GSettings.
-	 *     Need a mail_config_get_default_charset() that does this. */
-	if (!charset || charset[0] == '\0') {
+	if (!charset || !*charset) {
 		g_free (charset);
-		charset = g_settings_get_string (settings, "charset");
-		if (charset != NULL && *charset == '\0') {
-			g_free (charset);
-			charset = NULL;
-		}
+		charset = NULL;
 	}
 
 	g_object_unref (settings);
 
-	if (charset == NULL)
-		charset = g_strdup (camel_iconv_locale_charset ());
-
-	if (charset == NULL)
-		charset = g_strdup ("us-ascii");
+	if (!charset)
+		charset = g_strdup ("UTF-8");
 
 	return charset;
 }
@@ -518,7 +507,9 @@ e_composer_paste_html (EMsgComposer *composer,
 	g_return_val_if_fail (E_IS_MSG_COMPOSER (composer), FALSE);
 	g_return_val_if_fail (GTK_IS_CLIPBOARD (clipboard), FALSE);
 
-	html = e_clipboard_wait_for_html (clipboard);
+	if (!(html = e_clipboard_wait_for_html (clipboard)))
+		return FALSE;
+
 	g_return_val_if_fail (html != NULL, FALSE);
 
 	editor = e_msg_composer_get_editor (composer);
@@ -626,8 +617,8 @@ e_composer_paste_text (EMsgComposer *composer,
 	g_return_val_if_fail (E_IS_MSG_COMPOSER (composer), FALSE);
 	g_return_val_if_fail (GTK_IS_CLIPBOARD (clipboard), FALSE);
 
-	text = gtk_clipboard_wait_for_text (clipboard);
-	g_return_val_if_fail (text != NULL, FALSE);
+	if (!(text = gtk_clipboard_wait_for_text (clipboard)))
+		return FALSE;
 
 	editor = e_msg_composer_get_editor (composer);
 	view = e_html_editor_get_view (editor);
@@ -808,27 +799,6 @@ use_top_signature (EMsgComposer *composer)
 	return top_signature;
 }
 
-static void
-composer_size_allocate_cb (GtkWidget *widget,
-			   gpointer user_data)
-{
-	GtkWidget *scrolled_window;
-	GtkAdjustment *adj;
-
-	scrolled_window = gtk_widget_get_parent (GTK_WIDGET (widget));
-	adj = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (scrolled_window));
-
-	/* Scroll only when there is some size allocated */
-	if (gtk_adjustment_get_upper (adj) != 0.0) {
-		/* Scroll web view down to caret */
-		gtk_adjustment_set_value (adj, gtk_adjustment_get_upper (adj) - gtk_adjustment_get_page_size (adj));
-		gtk_scrolled_window_set_vadjustment (GTK_SCROLLED_WINDOW (scrolled_window), adj);
-		/* Disconnect because we don't want to scroll down the view on every window size change */
-		g_signal_handlers_disconnect_by_func (
-			widget, G_CALLBACK (composer_size_allocate_cb), NULL);
-	}
-}
-
 static WebKitDOMElement *
 prepare_paragraph (EHTMLEditorSelection *selection,
                    WebKitDOMDocument *document)
@@ -1046,15 +1016,10 @@ composer_move_caret (EMsgComposer *composer)
 		g_clear_object (&dom_selection);
 		g_clear_object (&dom_window);
 		g_clear_object (&range);
-
-		if (start_bottom)
-			e_html_editor_selection_scroll_to_caret (editor_selection);
 	}
 
 	if (start_bottom)
-		g_signal_connect (
-			view, "size-allocate",
-			G_CALLBACK (composer_size_allocate_cb), NULL);
+		e_html_editor_selection_scroll_to_caret (editor_selection);
 
 	e_html_editor_view_force_spell_check_in_viewport (view);
 
@@ -1066,20 +1031,20 @@ composer_load_signature_cb (EMailSignatureComboBox *combo_box,
                             GAsyncResult *result,
                             EMsgComposer *composer)
 {
-	GString *html_buffer = NULL;
 	gchar *contents = NULL;
 	gsize length = 0;
-	const gchar *active_id;
-	gboolean top_signature, is_html, html_mode;
-	gboolean start_bottom, is_message_from_edit_as_new;
+	gboolean top_signature, is_html, html_mode, is_message_from_edit_as_new;
 	GError *error = NULL;
+	gulong list_length, ii;
 	EHTMLEditor *editor;
 	EHTMLEditorView *view;
 	WebKitDOMDocument *document;
-	WebKitDOMElement *element = NULL;
+	WebKitDOMElement *signature_to_insert;
+	WebKitDOMElement *insert_signature_in = NULL;
+	WebKitDOMElement *signature_wrapper = NULL;
+	WebKitDOMElement *element, *converted_signature = NULL;
+	WebKitDOMHTMLElement *body;
 	WebKitDOMNodeList *signatures;
-	gulong list_length, ii;
-	GSettings *settings;
 
 	e_mail_signature_combo_box_load_selected_finish (
 		combo_box, result, &contents, &length, &is_html, &error);
@@ -1108,32 +1073,22 @@ composer_load_signature_cb (EMailSignatureComboBox *combo_box,
 		!is_message_from_edit_as_new &&
 		!composer->priv->is_from_new_message;
 
-	settings = e_util_ref_settings ("org.gnome.evolution.mail");
-	start_bottom = g_settings_get_boolean (settings, "composer-reply-start-bottom");
-	g_object_unref (settings);
-
+	/* Create the DOM signature that is the same across all types of signatures. */
 	document = webkit_web_view_get_dom_document (WEBKIT_WEB_VIEW (view));
+	body = webkit_dom_document_get_body (document);
+	signature_to_insert = webkit_dom_document_create_element (document, "span", NULL);
+	webkit_dom_element_set_class_name (signature_to_insert, "-x-evo-signature");
+	/* The combo box active ID is the signature's ESource UID. */
+	webkit_dom_element_set_id (
+		signature_to_insert, gtk_combo_box_get_active_id (GTK_COMBO_BOX (combo_box)));
+	insert_signature_in = signature_to_insert;
+
 	html_mode = e_html_editor_view_get_html_mode (view);
 
-	if (contents == NULL)
+	/* The signature has no content usually it means it is set to None. */
+	if (!contents)
 		goto insert;
 
-	/* If inserting HTML signature in plain text composer we have to convert it. */
-	if (is_html && !html_mode) {
-		WebKitDOMElement *tmp_element;
-		gchar *inner_text;
-
-		tmp_element = webkit_dom_document_create_element (document, "div", NULL);
-		webkit_dom_html_element_set_inner_html (
-			WEBKIT_DOM_HTML_ELEMENT (tmp_element), contents, NULL);
-		inner_text = webkit_dom_html_element_get_inner_text (
-			WEBKIT_DOM_HTML_ELEMENT (tmp_element));
-
-		g_free (contents);
-		contents = inner_text ? g_strstrip (inner_text) : g_strdup ("");
-		is_html = FALSE;
-	}
-
 	if (!is_html) {
 		gchar *html;
 
@@ -1144,22 +1099,31 @@ composer_load_signature_cb (EMailSignatureComboBox *combo_box,
 			contents = html;
 			length = strlen (contents);
 		}
+
+		insert_signature_in = webkit_dom_document_create_element (document, "pre", NULL);
+		webkit_dom_node_append_child (
+			WEBKIT_DOM_NODE (signature_to_insert),
+			WEBKIT_DOM_NODE (insert_signature_in),
+			NULL);
 	}
 
-	/* Generate HTML code for the signature. */
+	/* If inserting HTML signature in plain text composer we have to convert it. */
+	if (is_html && !html_mode && !strstr (contents, "data-evo-signature-plain-text-mode")) {
+		gchar *inner_text;
 
-	html_buffer = g_string_sized_new (1024);
+		/* Save the converted signature to avoid parsing it later again
+		 * while inserting it into the view. */
+		converted_signature = webkit_dom_document_create_element (document, "pre", NULL);
+		webkit_dom_html_element_set_inner_html (
+			WEBKIT_DOM_HTML_ELEMENT (converted_signature), contents, NULL);
+		e_html_editor_view_convert_element_from_html_to_plain_text (view, converted_signature);
+		inner_text = webkit_dom_html_element_get_inner_text (WEBKIT_DOM_HTML_ELEMENT (converted_signature));
 
-	/* The combo box active ID is the signature's ESource UID. */
-	active_id = gtk_combo_box_get_active_id (GTK_COMBO_BOX (combo_box));
-
-	g_string_append_printf (
-		html_buffer,
-		"<SPAN class=\"-x-evo-signature\" id=\"1\" name=\"%s\">",
-		(active_id != NULL) ? active_id : "");
-
-	if (!is_html)
-		g_string_append (html_buffer, "<PRE>");
+		g_free (contents);
+		contents = inner_text ? g_strstrip (inner_text) : g_strdup ("");
+		/* because of the -- \n check */
+		is_html = FALSE;
+	}
 
 	/* The signature dash convention ("-- \n") is specified
 	 * in the "Son of RFC 1036", section 4.3.2.
@@ -1182,16 +1146,37 @@ composer_load_signature_cb (EMailSignatureComboBox *combo_box,
 			;  /* skip */
 		else if (e_util_strstrcase (contents, delim_nl) != NULL)
 			;  /* skip */
-		else
-			g_string_append (html_buffer, delim);
+		else {
+			WebKitDOMElement *pre_delimiter;
+
+			pre_delimiter = webkit_dom_document_create_element (document, "pre", NULL);
+			/* Always use the HTML delimiter as we are never in anything
+			 * like a strict plain text mode. */
+			webkit_dom_html_element_set_inner_html (
+				WEBKIT_DOM_HTML_ELEMENT (pre_delimiter), "-- <br>", NULL);
+			webkit_dom_node_append_child (
+				WEBKIT_DOM_NODE (insert_signature_in),
+				WEBKIT_DOM_NODE (pre_delimiter),
+				NULL);
+		}
 	}
 
-	g_string_append_len (html_buffer, contents, length);
+	if (converted_signature) {
+		WebKitDOMNode *node;
 
-	if (!is_html)
-		g_string_append (html_buffer, "</PRE>");
+		while ((node = webkit_dom_node_get_first_child (WEBKIT_DOM_NODE (converted_signature))))
+			webkit_dom_node_append_child (
+				WEBKIT_DOM_NODE (insert_signature_in), node, NULL);
+		remove_node (WEBKIT_DOM_NODE (converted_signature));
+	} else
+		webkit_dom_html_element_insert_adjacent_html (
+			WEBKIT_DOM_HTML_ELEMENT (insert_signature_in), "beforeend", contents, NULL);
 
-	g_string_append (html_buffer, "</SPAN>");
+	element = webkit_dom_element_query_selector (
+		insert_signature_in, "[data-evo-signature-plain-text-mode]", NULL);
+	if (element)
+		webkit_dom_element_remove_attribute (
+			element, "data-evo-signature-plain-text-mode");
 	g_free (contents);
 
 insert:
@@ -1201,140 +1186,137 @@ insert:
 	list_length = webkit_dom_node_list_get_length (signatures);
 	for (ii = 0; ii < list_length; ii++) {
 		WebKitDOMNode *wrapper, *signature;
-		gchar *id;
 
 		wrapper = webkit_dom_node_list_item (signatures, ii);
 		signature = webkit_dom_node_get_first_child (wrapper);
 
+		/* Old messages will have the signature id in the name attribute, correct it. */
+		dom_element_rename_attribute (WEBKIT_DOM_ELEMENT (signature), "name", "id");
+
 		/* When we are editing a message with signature, we need to unset the
 		 * active signature id as if the signature in the message was edited
 		 * by the user we would discard these changes. */
-		if (composer->priv->set_signature_from_message &&
+		if (composer->priv->set_signature_from_message && contents &&
 		    (is_message_from_edit_as_new || e_html_editor_view_is_message_from_draft (view))) {
 			if (composer->priv->check_if_signature_is_changed) {
-				if (html_buffer && *html_buffer->str) {
-					gchar *body_signature_text, *signature_text;
-
-					element = webkit_dom_document_create_element (document, "div", NULL);
-					webkit_dom_html_element_set_inner_html (
-						WEBKIT_DOM_HTML_ELEMENT (element), html_buffer->str, NULL);
-
-					body_signature_text = webkit_dom_html_element_get_inner_text (
-						WEBKIT_DOM_HTML_ELEMENT (signature));
-					signature_text = webkit_dom_html_element_get_inner_text (
-						WEBKIT_DOM_HTML_ELEMENT (element));
-
+				/* Normalize the signature that we want to insert as the one in the
+				 * message already is normalized. */
+				webkit_dom_node_normalize (WEBKIT_DOM_NODE (signature_to_insert));
+				if (!webkit_dom_node_is_equal_node (WEBKIT_DOM_NODE (signature_to_insert), signature)) {
 					/* Signature in the body is different than the one with the
 					 * same id, so set the active signature to None and leave
 					 * the signature that is in the body. */
-					if (g_strcmp0 (body_signature_text, signature_text) != 0) {
-						gtk_combo_box_set_active (GTK_COMBO_BOX (combo_box), 0);
-						composer->priv->ignore_next_signature_change = TRUE;
-					}
-
-					g_free (body_signature_text);
-					g_free (signature_text);
-				} else {
-					gtk_combo_box_set_active (GTK_COMBO_BOX (combo_box), 0);
+					gtk_combo_box_set_active_id (GTK_COMBO_BOX (combo_box), "none");
 					composer->priv->ignore_next_signature_change = TRUE;
 				}
 
 				composer->priv->check_if_signature_is_changed = FALSE;
 				composer->priv->set_signature_from_message = FALSE;
 			} else {
-				gchar *name;
+				gchar *id;
 
 				/* Load the signature and check if is it the same
 				 * as the signature in body or the user previously
 				 * changed it. */
-				name = webkit_dom_element_get_attribute (WEBKIT_DOM_ELEMENT (signature), "name");
-				gtk_combo_box_set_active_id (GTK_COMBO_BOX (combo_box), name);
-				g_free (name);
+				id = webkit_dom_element_get_id (WEBKIT_DOM_ELEMENT (signature));
+				gtk_combo_box_set_active_id (GTK_COMBO_BOX (combo_box), id);
+				g_free (id);
 
 				composer->priv->check_if_signature_is_changed = TRUE;
 			}
 			g_object_unref (wrapper);
 			g_object_unref (signatures);
-
 			g_object_unref (composer);
 
 			return;
 		}
 
-		id = webkit_dom_element_get_id (WEBKIT_DOM_ELEMENT (signature));
-		if (id && (strlen (id) == 1) && (*id == '1')) {
-			/* If the top signature was set we have to remove the NL
-			 * that was inserted after it */
-			if (top_signature) {
-				WebKitDOMElement *spacer;
+		/* If the top signature was set we have to remove the newline
+		 * that was inserted after it */
+		if (top_signature) {
+			WebKitDOMElement *spacer;
 
-				spacer = webkit_dom_document_query_selector (
-					document, ".-x-evo-top-signature-spacer", NULL);
-				if (spacer)
-					remove_node_if_empty (WEBKIT_DOM_NODE (spacer));
-			}
-			/* We have to remove the div containing the span with signature */
-			remove_node (wrapper);
+			spacer = webkit_dom_document_query_selector (
+				document, ".-x-evo-top-signature-spacer", NULL);
+			if (spacer)
+				remove_node_if_empty (WEBKIT_DOM_NODE (spacer));
+		}
+		/* Leave just one signature wrapper there as it will be reused. */
+		if (ii != list_length - 1) {
 			g_object_unref (wrapper);
+			remove_node (wrapper);
+		} else {
+			remove_node (signature);
+			signature_wrapper = WEBKIT_DOM_ELEMENT (wrapper);
+		}
+	}
 
-			g_free (id);
-			break;
+	if (signature_wrapper) {
+		webkit_dom_node_append_child (
+			WEBKIT_DOM_NODE (signature_wrapper),
+			WEBKIT_DOM_NODE (signature_to_insert),
+			NULL);
+
+		/* Insert a spacer below the top signature */
+		if (top_signature && contents) {
+			WebKitDOMElement *spacer;
+			EHTMLEditorSelection *editor_selection;
+
+			editor_selection = e_html_editor_view_get_selection (view);
+			spacer = prepare_top_signature_spacer (editor_selection, document);
+			webkit_dom_node_insert_before (
+				WEBKIT_DOM_NODE (body),
+				WEBKIT_DOM_NODE (spacer),
+				webkit_dom_node_get_next_sibling (WEBKIT_DOM_NODE (signature_wrapper)),
+				NULL);
 		}
 
-		g_object_unref (wrapper);
-		g_free (id);
+		g_object_unref (signature_wrapper);
+	} else {
+		signature_wrapper = webkit_dom_document_create_element (document, "div", NULL);
+		webkit_dom_element_set_class_name (signature_wrapper, "-x-evo-signature-wrapper");
+
+		webkit_dom_node_append_child (
+			WEBKIT_DOM_NODE (signature_wrapper),
+			WEBKIT_DOM_NODE (signature_to_insert),
+			NULL);
+
+		if (top_signature) {
+			GSettings *settings;
+			WebKitDOMNode *child;
+
+			child = webkit_dom_node_get_first_child (WEBKIT_DOM_NODE (body));
+
+			settings = e_util_ref_settings ("org.gnome.evolution.mail");
+			if (g_settings_get_boolean (settings, "composer-reply-start-bottom")) {
+				webkit_dom_node_insert_before (
+					WEBKIT_DOM_NODE (body),
+					WEBKIT_DOM_NODE (signature_wrapper),
+					child,
+					NULL);
+			} else {
+				/* When we are using signature on top the caret
+				 * should be before the signature */
+				webkit_dom_node_insert_before (
+					WEBKIT_DOM_NODE (body),
+					WEBKIT_DOM_NODE (signature_wrapper),
+					child,
+					NULL);
+			}
+			g_object_unref (settings);
+		} else {
+			webkit_dom_node_append_child (
+				WEBKIT_DOM_NODE (body),
+				WEBKIT_DOM_NODE (signature_wrapper),
+				NULL);
+		}
+		composer_move_caret (composer);
 	}
 	g_object_unref (signatures);
 
-	if (html_buffer != NULL) {
-		if (*html_buffer->str) {
-			WebKitDOMHTMLElement *body;
-
-			body = webkit_dom_document_get_body (document);
-			if (!element) {
-				element = webkit_dom_document_create_element (document, "DIV", NULL);
-
-				webkit_dom_html_element_set_inner_html (
-					WEBKIT_DOM_HTML_ELEMENT (element), html_buffer->str, NULL);
-			}
-
-			webkit_dom_element_set_class_name (element, "-x-evo-signature-wrapper");
-
-			if (top_signature) {
-				WebKitDOMNode *child =
-					webkit_dom_node_get_first_child (WEBKIT_DOM_NODE (body));
-
-				if (start_bottom) {
-					webkit_dom_node_insert_before (
-						WEBKIT_DOM_NODE (body),
-						WEBKIT_DOM_NODE (element),
-						child,
-						NULL);
-				} else {
-					/* When we are using signature on top the caret
-					 * should be before the signature */
-					webkit_dom_node_insert_before (
-						WEBKIT_DOM_NODE (body),
-						WEBKIT_DOM_NODE (element),
-						child,
-						NULL);
-				}
-			} else {
-				webkit_dom_node_append_child (
-					WEBKIT_DOM_NODE (body),
-					WEBKIT_DOM_NODE (element),
-					NULL);
-			}
-		}
-
-		g_string_free (html_buffer, TRUE);
-	}
-
 	if (is_html && html_mode)
 		e_html_editor_view_fix_file_uri_images (view);
 
-	composer_move_caret (composer);
-
  exit:
 	/* Make sure the flag will be unset and won't influence user's choice */
 	composer->priv->set_signature_from_message = FALSE;
